#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include "filter.h"
#include <libspe2.h>
#include <pthread.h>
#include <ppu_intrinsics.h>

// Anzahl der SPUs, die verwendet werden sollen
#define SPE_THREADS 4

// Programm-Handle, um das SPU-Programm zu laden
extern spe_program_handle_t filter_spu;

// Filter-Koeffizienten
float a0 = 0.5, a1 = - 0.5;

// Struktur zur Datenbergaben an die PPU-Threads
typedef struct ppu_thread_data
{
  spe_context_ptr_t spe_ctx;  // SPU-Kontext
  pthread_t pthread;          // Das Thread-Objekt selbst
  void *argp;                 // Pointer, der auf einen parm_context zeigt (siehe filter.h)
} ppu_thread_data_t;

// Funktion, die der Thread ausfhrt
void *ppu_pthread_function(void *arg)
{
  // In argp wird ein Pointer auf die ppu_thread_data-Struktur bergeben
  ppu_thread_data_t *datap = (ppu_thread_data_t *)arg;

  unsigned int entry = SPE_DEFAULT_ENTRY;  // gibt an, dass main() im SPU-Programm ausgefhrt werden soll

  // Aufruf des SPU-Programms.
  // 1. Argument:  SPU-Kontext, um die richtige SPU zu finden
  // 2. Argument:  SPU-lokale Start-Adresse oder SPE_DEFAULT_ENTRY fr main()
  // 3. Argument:  Flags.  0 = default behaviour
  // 4. Argument:  Optionaler Pointer, der an die SPU bergeben wird (als zweites Argument)
  // 5. Argument:  ein weiterer solcher Pointer (als drittes Argument bergeben)
  // 6. Argument:  stopinfo-Pointer falls Info ber SPU-Beendigungsgrund bentigt wird
  if (spe_context_run (datap->spe_ctx, &entry, 0, datap->argp, NULL, NULL) < 0)
  {
    perror ("Failed running context");
    exit (1);
  }

  // Thread beenden
  pthread_exit(NULL);
}


int main (int argc, char *argv[])
{
  if (argc != 3) { fputs ("usage: filter length iterations\n", stderr);  exit (1); }
  int size = atoi (argv[1]);  // Lnge des zu filternden Arrays
  int iter = atoi (argv[2]);  // Anzahl der Wiederholungen (zur besseren Zeitmessung)

  float *m = calloc (size+2, sizeof (float));  // Quell-Array
  float *x = m + 2;
  float *y = calloc (size, sizeof (float));  // Ziel-Array
  float *y2;

  // Koeffizienten-Array zur bergabe an SPU
  float a[] __attribute__ ((aligned (16))) = {a0, a1};

  int it, n, i;

  // zur Zeitmessung:
  struct timeval T1, T2, T3, T4;

  // ein Thread-Objekt fr jede SPU
  ppu_thread_data_t datas[SPE_THREADS];

  // ein Parameter-Kontext fr jede SPU
  parm_context ctxs[SPE_THREADS] __attribute__ ((aligned (16)));

  // Zielarray allozieren
  posix_memalign ((void **)&y2, 4 * sizeof (float), (size+4) * sizeof (float));

  // Quelldaten initialisieren
  for (n = 0; n < size; n ++)  x[n] = n % 13;

#if 0
  // Versuch, Quell-Array aus dem Cache zu flushen
  { void *p = (void *) x;
    for (n = 0; n < size/128; n ++)
    { __dcbf (p);
      p += 128;
    }
  }
#endif

  // SPUs initialisieren
  for (n = 0; n < SPE_THREADS; n ++)
  {
      // SPU-Kontext fr jede SPU erzeugen
      if ((datas[n].spe_ctx = spe_context_create (0, NULL)) == NULL) {
          perror ("Failed creating context");
          exit (1);
      }

      // SPU-Programm in jede SPU laden
      if (spe_program_load (datas[n].spe_ctx, &filter_spu)) {
        perror ("Failed loading program");
        exit (1);
      }
  }

  gettimeofday (&T1, NULL);  // Zeitmessung Beginn sequentieller Algorithmus

  for (it = 0; it <= iter; it ++)
  {
    if (it == 1)  gettimeofday (&T1, NULL);
    for (n = 0; n < size; n ++)
      y[n] = x[n] * a0 + x[n-1] * a1;
  }

  gettimeofday (&T2, NULL);  // Zeitmessung Ende sequentieller Algorithmus

  gettimeofday (&T3, NULL);  // wird nach 1. Iteration upgedatet

  for (it = 0; it <= iter; it ++)  // iter+1 Iterationen, da 1. nicht gemessen wird
  {
    // Zeitmessung SPU-Algorithmus: Beginn nach 1. Iteration
    if (it == 1)  gettimeofday (&T3, NULL);

    for (n = 0; n < SPE_THREADS; n ++)
    {
      // Parameter-Kontext fr SPUs zusammenstellen
      ctxs[n].size = size / SPE_THREADS;  // zu bearbeitende Block-Gre
      ctxs[n].x = x + n * ctxs[n].size;   // Pointer auf Quell-Block
      ctxs[n].y = y2 + n * ctxs[n].size;  // Pointer auf Ziel-Block
      ctxs[n].a = a;                      // Pointer auf Koeffizienten

      // Parameter-Kontext in Thread-Kontext eintragen
      datas[n].argp = &ctxs[n];

      // Threads erzeugen und starten
      if (pthread_create (&datas[n].pthread, NULL, &ppu_pthread_function, &datas[n])) {
        perror ("Failed creating thread");
      }
    }

    // Warten, bis alle Threads fertig sind
    for (n = 0; n < SPE_THREADS; n ++)
    {
      if (pthread_join (datas[n].pthread, NULL))
      { perror ("Failed joining thread");
        exit (1);
      }
    }
  }

  gettimeofday (&T4, NULL);  // Zeitmessung End SPU-Algorithmus

  // Unterschied zwischen den zwei Algorithmen berechnen, um Fehler zu erkennen
  double error = 0.0;
  for (i = 0; i < size; i ++)
  {
    double diff = y[i] - y2[i];
    error += diff * diff;
  }
  error /= (double) size;

  // Ausgabe von Ausfhrungszeiten, Speedup und Fehler
  int DT1 = (T2.tv_sec - T1.tv_sec) * 1000 + (T2.tv_usec - T1.tv_usec) / 1000;
  int DT2 = (T4.tv_sec - T3.tv_sec) * 1000 + (T4.tv_usec - T3.tv_usec) / 1000;
  printf ( " plain time: %d ms\n vector time: %d ms\n speedup: %f\n error: %f\n"
         , DT1, DT2, (double) DT1 / (double) DT2
         , error
         );

  // Speicher freigeben
  free (m);
  free (y);
  free (y2);

  return 0;
}
